unit DBBuilder.Core.Config;

interface

uses
  System.SysUtils, System.Classes, System.IOUtils, DBBuilder.Core.IniPersistance;

type
  TDatabaseInfo = class
  private
    FPort: Integer;
    FDatabase: string;
    FServer: string;
    FCharset: string;
    FPassword: string;
    FUserName: string;
    FClientLib: string;
    FEngine: string;
    function GetConnectionString: string;
  public
    [IniPersistSkip]
    property ConnectionString: string read GetConnectionString;
    property Engine: string read FEngine write FEngine;
    property Server: string read FServer write FServer;
    property Port: Integer read FPort write FPort;
    property Database: string read FDatabase write FDatabase;
    property UserName: string read FUserName write FUserName;
    property Password: string read FPassword write FPassword;
    property ClientLib: string read FClientLib write FClientLib;
    property Charset: string read FCharset write FCharset;
  end;

  TConfig = class
  private
    FScriptPath: string;
    FAppName: string;
    FAppPath: string;
    FTargetDatabase: TDatabaseInfo;
    FSourceDatabase: TDatabaseInfo;
    FIniFileName: string;
    FMapPath: string;
    procedure InitConfig;
  public
    property SourceDatabase: TDatabaseInfo read FSourceDatabase;
    property TargetDatabase: TDatabaseInfo read FTargetDatabase;
    [IniPersistSkip]
    property ScriptPath: string read FScriptPath;
    [IniPersistSkip]
    property MapPath: string read FMapPath;
    [IniPersistSkip]
    property AppName: string read FAppName;
    [IniPersistSkip]
    property AppPath: string read FAppPath;
    [IniPersistSkip]
    property IniFileName: string read FIniFileName;

    procedure Read;
    procedure Write;

    constructor Create;
    destructor Destroy; override;
  end;

function Config: TConfig;

var
  DefaultIniFileName: string = 'DBBuilder.ini';

implementation

var
  _Config: TConfig;

function Config: TConfig;
begin
  if not Assigned(_Config) then
  begin
    _Config := TConfig.Create;
    _Config.Read;
    if not FileExists(_Config.IniFileName) then
      _Config.Write;
  end;
  Result := _Config;
end;

{ TConfig }

constructor TConfig.Create;
begin
  FAppName := ParamStr(0);
  FAppPath := ExtractFilePath(FAppName);
  FIniFileName := ExtractFilePath(ParamStr(0)) + DefaultIniFileName;
  FScriptPath := TPath.Combine(FAppPath, 'scripts');
  FMapPath := TPath.Combine(FAppPath, 'maps');
  ForceDirectories(FScriptPath);

  FSourceDatabase := TDatabaseInfo.Create;
  FTargetDatabase := TDatabaseInfo.Create;

  InitConfig;
end;

destructor TConfig.Destroy;
begin
  FSourceDatabase.Free;
  FTargetDatabase.Free;
  inherited;
end;

procedure TConfig.InitConfig;
begin
  FSourceDatabase.Engine := 'FDPostgres';
  FSourceDatabase.Server := 'localhost';
  FSourceDatabase.Port := 5432;
  FSourceDatabase.Database := 'test';
  FSourceDatabase.UserName := 'postgres';
  FSourceDatabase.Password := 'postgres';
  FSourceDatabase.ClientLib := 'C:\Program Files (x86)\SQL Maestro Group\PostgreSQL Maestro\libpq.dll';

  FTargetDatabase.Engine := 'FDFirebird';
  FTargetDatabase.Server := 'localhost';
  FTargetDatabase.Port := 3052;
  FTargetDatabase.Database := 'c:\db\ingenia.fdb';
  FTargetDatabase.UserName := 'SYSDBA';
  FTargetDatabase.Password := 'masterkey';
  FTargetDatabase.ClientLib := 'C:\Firebird\Firebird-4.0.0.1436-0\fbclient.dll';
  FTargetDatabase.Charset := 'ISO8859_1';
end;

procedure TConfig.Read;
begin
  TIniPersistent.Load(FIniFileName, Self);
end;

procedure TConfig.Write;
begin
  TIniPersistent.Save(FIniFileName, Self);
end;

{ TDatabaseInfo }

function TDatabaseInfo.GetConnectionString: string;
begin
  Result := Format('%s://%s@%s:%d/%s', [FEngine, FUserName, FServer, FPort, FDatabase])
end;

initialization

  _Config := nil;

finalization

  _Config.Free;

end.
