unit DBBuilder.Core.IniPersistance;

interface

uses
  System.Classes, System.SysUtils, System.Rtti, System.IniFiles,
  System.Generics.Collections, Winapi.Windows;

type
  IniPersistSkipAttribute = class(TCustomAttribute)
  end;

  IniPersistNameAttribute = class(TCustomAttribute)
  private
    FPropertyName: string;
  public
    property PropertyName: string read FPropertyName write FPropertyName;
    constructor Create(const APropertyName: string);
  end;

  IniPersistEnumAttribute = class(TCustomAttribute)
  private
    FValueList: TArray<string>;
  public
    function GetValue(Index: Integer): string;
    function GetIndex(const Value: string): Integer;
    property ValueList: TArray<string> read FValueList write FValueList;
    constructor Create(const AValueList: string);
  end;

  TIniPersistent = class(TObject)
  private
    const
      DefaultSectionName = 'General';
    class procedure ReadSection(IniFile: TIniFile; const SectionName: string; RTTIType: TRttiType; Obj: TObject); static;
    class procedure WriteSection(IniFile: TIniFile; const SectionName: string; RTTIType: TRttiType; Obj: TObject); static;
  public
    class procedure Load(const Filename: string; Obj: TObject);
    class procedure Save(const Filename: string; Obj: TObject);
  end;

implementation

uses
  DBBuilder.Core.Rtti;

{ TIniPersistent }

class procedure TIniPersistent.ReadSection(IniFile: TIniFile; const SectionName: string; RTTIType: TRttiType; Obj: TObject);
var
  Prop: TRttiProperty;
  //EncryptAttr: ELPersistEncryptAttribute;
  PropNameAttr: IniPersistNameAttribute;
  EnumAttr: IniPersistEnumAttribute;
  PropName: string;
begin
  for Prop in RTTIType.GetProperties do
  begin
    try
      if TRttiUtils.HasAttribute<IniPersistSkipAttribute>(Prop) then
        Continue;
      PropNameAttr := TRttiUtils.FindAttribute<IniPersistNameAttribute>(Prop);
      if Assigned(PropNameAttr) then
        PropName := PropNameAttr.PropertyName
      else
        PropName := Prop.Name;

      case Prop.PropertyType.TypeKind of
        tkString, tkUnicodeString:
        begin
//            EncryptAttr := TRttiUtils.FindAttribute<ELPersistEncryptAttribute>(Prop);
//            if Assigned(EncryptAttr) then
//              Prop.SetValue(Obj, IniFile.ReadEncString(SectionName, PropName, Prop.GetValue(Obj).AsString, EncryptAttr.Password, VigenerePassLen, EncryptAttr.EncryptMethod))
//            else
            Prop.SetValue(Obj, IniFile.ReadString(SectionName, PropName, Prop.GetValue(Obj).AsString));
        end;
        tkInteger:
          Prop.SetValue(Obj, IniFile.ReadInteger(SectionName, PropName, Prop.GetValue(Obj).AsInteger));
        tkFloat:
        begin
          if Prop.GetValue(Obj).TypeInfo = TypeInfo(TDate) then
            Prop.SetValue(Obj, StrToDate(IniFile.ReadString(SectionName, PropName, DateToStr(Prop.GetValue(Obj).AsExtended))))
          else if Prop.GetValue(Obj).TypeInfo = TypeInfo(TDateTime) then
            Prop.SetValue(Obj, StrToDateTime(IniFile.ReadString(SectionName, PropName, DateTimeToStr(Prop.GetValue(Obj).AsExtended))))
          else if Prop.GetValue(Obj).TypeInfo = TypeInfo(TTime) then
            Prop.SetValue(Obj, StrToTime(IniFile.ReadString(SectionName, PropName, TimeToStr(Prop.GetValue(Obj).AsExtended))))
          else
            Prop.SetValue(Obj, IniFile.ReadFloat(SectionName, PropName, Prop.GetValue(Obj).AsExtended));
        end;
        tkEnumeration:
        begin
          EnumAttr := TRttiUtils.FindAttribute<IniPersistEnumAttribute>(Prop);
          if Prop.PropertyType.Handle = TypeInfo(Boolean) then
            Prop.SetValue(Obj, IniFile.ReadInteger(SectionName, PropName, Prop.GetValue(Obj).AsOrdinal) > 0)
          else if Assigned(EnumAttr) then
          begin
            Prop.SetValue(Obj,
              TValue.FromOrdinal(
                Prop.PropertyType.Handle,
                EnumAttr.GetIndex(
                  IniFile.ReadString(SectionName, PropName, EnumAttr.GetValue(Prop.GetValue(Obj).AsOrdinal))
                )
              )
            );
          end
          else
            Prop.SetValue(Obj, IniFile.ReadInteger(SectionName, PropName, Prop.GetValue(Obj).AsOrdinal));
        end;
        tkClass:
        begin
          if SectionName <> DefaultSectionName then
            raise Exception.Create('Inner section not supported');

          ReadSection(IniFile, PropName, Prop.PropertyType, Prop.GetValue(Obj).AsObject);
        end;
      end;
    except
      on E: Exception do
        OutputDebugString(PChar('IniPersistent: ' + E.Message));
    end;
  end;
end;


class procedure TIniPersistent.Load(const Filename: string; Obj: TObject);
var
  IniFile: TIniFile;
  RTTIContext: TRttiContext;
  RTTIType: TRttiType;
begin
  IniFile := TIniFile.Create(Filename);
  try
    RTTIContext := TRttiContext.Create;
    try
      RTTIType := RTTIContext.GetType(Obj.ClassInfo);
      ReadSection(IniFile, DefaultSectionName, RTTIType, Obj);
    finally
      RTTIContext.Free;
    end;
  finally
    IniFile.Free;
  end;
end;

class procedure TIniPersistent.WriteSection(IniFile: TIniFile; const SectionName: string; RTTIType: TRttiType; Obj: TObject);
var
  Prop: TRttiProperty;
  //EncryptAttr: IniPersistEncryptAttribute;
  PropNameAttr: IniPersistNameAttribute;
  EnumAttr: IniPersistEnumAttribute;
  PropName: string;
begin
  for Prop in RTTIType.GetProperties do
  begin
    if TRttiUtils.HasAttribute<IniPersistSkipAttribute>(Prop) then
      Continue;
    PropNameAttr := TRttiUtils.FindAttribute<IniPersistNameAttribute>(Prop);
    if Assigned(PropNameAttr) then
      PropName := PropNameAttr.PropertyName
    else
      PropName := Prop.Name;
    case Prop.PropertyType.TypeKind of
      tkString, tkUnicodeString:
      begin
//          EncryptAttr := TRttiUtils.FindAttribute<ELPersistEncryptAttribute>(Prop);
//          if Assigned(EncryptAttr) then
//            IniFile.WriteEncString(SectionName, PropName, Prop.GetValue(Obj).AsString, EncryptAttr.Password, VigenerePassLen, EncryptAttr.EncryptMethod)
//          else
          IniFile.WriteString(SectionName, PropName, Prop.GetValue(Obj).AsString);
      end;
      tkInteger:
        IniFile.WriteInteger(SectionName, PropName, Prop.GetValue(Obj).AsInteger);
      tkFloat:
      begin
        if Prop.GetValue(Obj).TypeInfo = TypeInfo(TDate) then
          IniFile.WriteString(SectionName, PropName, DateToStr(Prop.GetValue(Obj).AsExtended))
        else if Prop.GetValue(Obj).TypeInfo = TypeInfo(TDateTime) then
          IniFile.WriteString(SectionName, PropName, DateTimeToStr(Prop.GetValue(Obj).AsExtended))
        else if Prop.GetValue(Obj).TypeInfo = TypeInfo(TTime) then
          IniFile.WriteString(SectionName, PropName, TimeToStr(Prop.GetValue(Obj).AsExtended))
        else
          IniFile.WriteFloat(SectionName, PropName, Prop.GetValue(Obj).AsExtended);
      end;
      tkEnumeration:
      begin
        EnumAttr := TRttiUtils.FindAttribute<IniPersistEnumAttribute>(Prop);
        if Prop.PropertyType.Handle = TypeInfo(Boolean) then
          IniFile.WriteInteger(SectionName, PropName, Prop.GetValue(Obj).AsOrdinal)
        else if Assigned(EnumAttr) then
          IniFile.WriteString(SectionName, PropName, EnumAttr.GetValue(Prop.GetValue(Obj).AsOrdinal))
        else
          IniFile.WriteInteger(SectionName, PropName, Prop.GetValue(Obj).AsOrdinal);
      end;
      tkClass:
      begin
        if SectionName <> DefaultSectionName then
          raise Exception.Create('Inner section not supported');

        WriteSection(IniFile, PropName, Prop.PropertyType, Prop.GetValue(Obj).AsObject);
      end;
    end;
  end;
end;

class procedure TIniPersistent.Save(const Filename: string; Obj: TObject);
var
  IniFile: TIniFile;
  RTTIContext: TRttiContext;
  RTTIType: TRttiType;
begin
  IniFile := TIniFile.Create(Filename);
  try
    RTTIContext := TRttiContext.Create;
    try
      RTTIType := RTTIContext.GetType(Obj.ClassInfo);
      WriteSection(IniFile, DefaultSectionName, RTTIType, Obj);
    finally
      RTTIContext.Free;
    end;
  finally
    IniFile.Free;
  end;end;

{ IniPersistNameAttribute }

constructor IniPersistNameAttribute.Create(const APropertyName: string);
begin
  inherited Create;
  FPropertyName := APropertyName;
end;

{ IniPersistEnumAttribute }

constructor IniPersistEnumAttribute.Create(const AValueList: string);
begin
  inherited Create;
  FValueList := AValueList.Split([',']);
end;

function IniPersistEnumAttribute.GetIndex(const Value: string): Integer;
begin
  if not TArray.BinarySearch<string>(FValueList, Value, Result) then
    Result := 0;
end;

function IniPersistEnumAttribute.GetValue(Index: Integer): string;
begin
  if Index < Length(FValueList) then
    Result := FValueList[Index]
  else
    Result := IntToStr(Index);
end;

end.
