unit DBBuilder.Core.Lexer;

interface

uses
  System.Classes, System.SysUtils, System.Math, Generics.Collections;

type
  {$SCOPEDENUMS ON}
  TTokenKind = (
    EndOfStream,
    StringLiteral,
    IntegerLiteral,
    FloatLiteral,
    Identifier,
    Plus,
    Minus,
    Mult,
    Divide,
    Power,
    Assignment,
    Semicolon,
    LeftParenthesis,
    RightParenthesis,
    LeftSquareBracket,
    RightSquareBracket,
    LeftCurleyBracket,
    RightCurleyBracket,
    Comma,
    KeywordIf,
    KeywordEnd,
    KeywordPrintln,
    KeywordThen,
    KeywordUntil,
    KeywordElse,
    KeywordWhile,
    KeywordFor,
    KeywordRepeat,
    KeywordReturn,
    KeywordBreak,
    KeywordFunction,
    KeywordDo,
    KeywordTo,
    KeywordDownto,
    KeywordRef,
    KeywordNot,
    KeywordFalse,
    KeywordTrue,
    Equivalence,
    NotEqual,
    LessThen,
    LessThenOrEqual,
    MoreThen,
    MoreThenOrEqual,
    IdentifierNot,
    BinaryAnd,
    LogicalAnd,
    BinaryOr,
    LogicalOr,
    Hash,
    KeywordNull
  );
  {$SCOPEDENUMS OFF}

  ELexerError = class(Exception)
  private
    FCol: Integer;
    FLine: Integer;
  public
    property Line: Integer read FLine;
    property Col: Integer read FCol;
    constructor Create(const AMessage: string; ALine, ACol: Integer);
  end;

  TToken = record
    LineNumber: Integer;
    ColumnNumber: Integer;
    Kind: TTokenKind;
    StringValue: string;
    FloatValue: Double;
    IntegerValue: Integer;
    function ToString: string;
  end;

  TBufferReader = class(TObject)
  strict private
    FStreamReader: TStreamReader;
    FColumnNumber: Integer;
    FLineNumber: Integer;
    FBuffer: string;
    FBufferIndex: Integer;
    function ReadRawChar: Char;
    procedure StartScan;
  public
    const EOF_CHAR = #$ffff;
    const CR = #$0d;
    const LF = #$0a;
    const TAB = #$09;

    function ReadChar: Char;
    function ReadNextChar(Position: Integer = 1): Char;

    property ColumnNumber: Integer read FColumnNumber;
    property LineNumber: Integer read FLineNumber;

    constructor Create(AStream: TStream; AEncoding: TEncoding; AOwnStream: Boolean);
    destructor Destroy; override;
  end;

  TTokenScanner = class(TObject)
  private
    FBufferReader: TBufferReader;
    FCurrentChar: Char;
    FKeywords: TDictionary<string, TTokenKind>;

    procedure InitKeywords;
    function IsKeyword(const AIdentifier: string; out ATokenKind: TTokenKind): Boolean;

    procedure SkipBlankAndComment;
    procedure SkipSingleLineComment;
    procedure SkipMultiLineComment;

    function GetWord: TToken;
    function GetString: TToken;
    function GetNumber: TToken;
    function GetSpecial: TToken;
    function GetMultiLineString: TToken;
  public
    function NextToken: TToken;

    constructor CreateFromFile(const AFileName: TFileName; AEncoding: TEncoding);
    constructor CreateFromString(const AValue: string);
    constructor CreateFromStream(AStream: TStream; AEncoding: TEncoding; AOwnStream: Boolean);
    destructor Destroy; override;
  end;

function KindToString(Kind: TTokenKind): string;

implementation

{ Utils }

const
  MAX_EXPONENT = 308;

function KindToString(Kind: TTokenKind): string;
const
  KindStr: array [TTokenKind] of string = (
    'EOF',                 //  EndOfStream,
    'String',              //  StringLiteral,
    'Integer',             //  IntegerLiteral,
    'Double',              //  FloatLiteral,
    'Identifier',          //  Identifier,
    'Plus',                //  Plus,
    'Minus',               //  Minus,
    'Mult',                //  Mult,
    'Divide',              //  Divide,
    'Power',               //  Power,
    'Assignment',          //  Assignment,
    'Semicolon',           //  Semicolon,
    'LeftParenthesis',     //  LeftParenthesis,
    'RightParenthesis',    //  RightParenthesis,
    'LeftSquareBracket',   //  LeftSquareBracket,
    'RightSquareBracket',  //  RightSquareBracket,
    'LeftCurleyBracket',   //  LeftCurleyBracket,
    'RightCurleyBracket',  //  RightCurleyBracket,
    'Comma',               //  Comma,
    'KeywordIf',           //  KeywordIf,
    'KeywordEnd',          //  KeywordEnd,
    'KeywordPrintln',      //  KeywordPrintln,
    'KeywordThen',         //  KeywordThen,
    'KeywordUntil',        //  KeywordUntil,
    'KeywordElse',         //  KeywordElse,
    'KeywordWhile',        //  KeywordWhile,
    'KeywordFor',          //  KeywordFor,
    'KeywordRepeat',       //  KeywordRepeat,
    'KeywordReturn',       //  KeywordReturn,
    'KeywordBreak',        //  KeywordBreak,
    'KeywordFunction',     //  KeywordFunction,
    'KeywordDo',           //  KeywordDo,
    'KeywordTo',           //  KeywordTo,
    'KeywordDownto',       //  KeywordDownto,
    'KeywordRef',          //  KeywordRef,
    'KeywordNot',          //  KeywordNot,
    'KeywordFalse',        //  KeywordFalse,
    'KeywordTrue',         //  KeywordTrue,
    'Equivalence',         //  Equivalence,
    'NotEqual',            //  NotEqual,
    'LessThen',            //  LessThen,
    'LessThenOrEqual',     //  LessThenOrEqual,
    'MoreThen',            //  MoreThen,
    'MoreThenOrEqual',     //  MoreThenOrEqual,
    'Not',                 //  IdentifierNot,
    'BinaryAnd',           //  BinaryAnd,
    'LogicalAnd',          //  LogicalAnd,
    'BinaryOr',            //  BinaryOr,
    'LogicalOr',           //  LogicalOr
    'Hash',                //  Hash (#)
    'KeywordNull'          //  KeywordNull
  );
begin
  Result := KindStr[Kind];
end;

function KindToType(Kind: TTokenKind): Integer;
const
  KindType: array [TTokenKind] of Integer = (
    varUnknown,           //  EndOfStream,
    varUString,           //  StringLiteral,
    varInteger,           //  IntegerLiteral,
    varDouble,            //  FloatLiteral,
    varUString,           //  Identifier,
    varUnknown,           //  Plus,
    varUnknown,           //  Minus,
    varUnknown,           //  Mult,
    varUnknown,           //  Divide,
    varUnknown,           //  Power,
    varUnknown,           //  Assignment,
    varUnknown,           //  Semicolon,
    varUnknown,           //  LeftParenthesis,
    varUnknown,           //  RightParenthesis,
    varUnknown,           //  LeftSquareBracket,
    varUnknown,           //  RightSquareBracket,
    varUnknown,           //  LeftCurleyBracket,
    varUnknown,           //  RightCurleyBracket,
    varUnknown,           //  Comma,
    varUString,           //  KeywordIf,
    varUString,           //  KeywordEnd,
    varUString,           //  KeywordPrintln,
    varUString,           //  KeywordThen,
    varUString,           //  KeywordUntil,
    varUString,           //  KeywordElse,
    varUString,           //  KeywordWhile,
    varUString,           //  KeywordFor,
    varUString,           //  KeywordRepeat,
    varUString,           //  KeywordReturn,
    varUString,           //  KeywordBreak,
    varUString,           //  KeywordFunction,
    varUString,           //  KeywordDo,
    varUString,           //  KeywordTo,
    varUString,           //  KeywordDownto,
    varUString,           //  KeywordRef,
    varUString,           //  KeywordNot,
    varUString,           //  KeywordFalse,
    varUString,           //  KeywordTrue,
    varUnknown,           //  Equivalence,
    varUnknown,           //  NotEqual,
    varUnknown,           //  LessThen,
    varUnknown,           //  LessThenOrEqual,
    varUnknown,           //  MoreThen,
    varUnknown,           //  MoreThenOrEqual,
    varUnknown,           //  IdentifierNot,
    varUnknown,           //  BinaryAnd,
    varUnknown,           //  LogicalAnd,
    varUnknown,           //  BinaryOr,
    varUnknown,           //  LogicalOr
    varUnknown,           //  Hash
    varUString            //  KeywordNull,
  );
begin
  Result := KindType[Kind];
end;

function IsBlank(const AChar: Char): Boolean; inline;
begin
  Result := CharInSet(AChar, [' ', TBufferReader.TAB, TBufferReader.CR, TBufferReader.LF]);
end;

function IsLetter(const AChar: Char): Boolean; inline;
begin
  Result := CharInSet(AChar, ['a'..'z', 'A'..'Z', '_']);
end;

function IsDigit(const AChar: Char): Boolean; inline;
begin
  Result := CharInSet(AChar, ['0'..'9', '.']);
end;

function IsInteger(const AChar: Char): Boolean; inline;
begin
  Result := CharInSet(AChar, ['0'..'9']);
end;

function IsEof(const AChar: Char): Boolean; inline;
begin
  Result := AChar = TBufferReader.EOF_CHAR;
end;

{ TBufferReader }

{ TBufferReader }

destructor TBufferReader.Destroy;
begin
  FreeAndNil(FStreamReader);
  inherited;
end;

function TBufferReader.ReadChar: Char;
begin
  if FBufferIndex > 0 then
  begin
    Dec(FBufferIndex);
    Exit(FBuffer[FBufferIndex + 1]);
  end;

  Result := ReadRawChar;
  if Result = LF then
  begin
    Inc(FLineNumber);
    FColumnNumber := 0;
    //Result := ' ';
  end;
end;

function TBufferReader.ReadNextChar(Position: Integer = 1): Char;
begin
  ///Result := EOF_CHAR;
  SetLength(FBuffer, FBufferIndex);
  while FBufferIndex < Position do
  begin
    FBuffer := FBuffer + ReadRawChar;
    Inc(FBufferIndex);
  end;
  Result := FBuffer[Position];
end;

function TBufferReader.ReadRawChar: Char;

  function InternalReadChar: Char;
  begin
    if not Assigned(FStreamReader) then
    begin
      Exit(EOF_CHAR);
    end;

    if FStreamReader.EndOfStream then
    begin
      FreeAndNil(FStreamReader);
      Exit(EOF_CHAR);
    end;

    Inc(FColumnNumber);
    Result := Char(FStreamReader.Read);
  end;

var
  LCurrentChar: Char;
begin
  LCurrentChar := InternalReadChar;
  if (LCurrentChar = CR) or (LCurrentChar = LF) then
  begin
    if LCurrentChar = CR then
    begin
      LCurrentChar := InternalReadChar;
      if LCurrentChar = LF then
        Exit(LCurrentChar);
      raise ELexerError.Create('Expecting line feed character', FLineNumber, FColumnNumber);
    end;
    Exit(LCurrentChar);
  end
  else
    Result := LCurrentChar;
end;

constructor TBufferReader.Create(AStream: TStream; AEncoding: TEncoding; AOwnStream: Boolean);
begin
  FStreamReader := TStreamReader.Create(AStream, AEncoding);
  if AOwnStream then
    FStreamReader.OwnStream;

  StartScan;
end;

procedure TBufferReader.StartScan;
begin
  FLineNumber := 1;
  FColumnNumber := 0;
end;

{ TTokenScanner }

destructor TTokenScanner.Destroy;
begin
  FKeywords.Free;
  FBufferReader.Free;
  inherited;
end;

function TTokenScanner.GetNumber: TToken;
var
  HasLeftSide, HasRightSide: Boolean;
  SingleDigit: Integer;
  Scale: Double;
  ExponentSign: Integer;
  Evalue: Integer;
begin
  HasLeftSide := False;
  HasRightSide := False;

  Result.IntegerValue := 0;
  Result.FloatValue := 0.0;

  Result.Kind := TTokenKind.IntegerLiteral;

  if FCurrentChar <> '.' then
  begin
    HasLeftSide := True;
    repeat
      SingleDigit := Ord(FCurrentChar) - Ord('0');
      if Result.IntegerValue > (MaxInt - SingleDigit) div 10 then
        raise ELexerError.Create('Integer overflow', FBufferReader.LineNumber, FBufferReader.ColumnNumber);

      Result.IntegerValue := 10 * Result.IntegerValue + SingleDigit;
      FCurrentChar := FBufferReader.ReadChar;

    until not IsInteger(FCurrentChar);
  end;

  Scale := 1;
  if FCurrentChar = '.' then
  begin
    Result.Kind := TTokenKind.FloatLiteral;
    Result.FloatValue := Result.IntegerValue;
    FCurrentChar := FBufferReader.ReadChar;
    if IsInteger(FCurrentChar) then
      HasRightSide := True;
    while IsDigit(FCurrentChar) do
    begin
      Scale := Scale * 0.1;
      SingleDigit := Ord(FCurrentChar) - Ord('0');
      Result.FloatValue := Result.FloatValue + (SingleDigit * Scale);
      FCurrentChar := FBufferReader.ReadChar;
    end;

  end;

  if (not HasLeftSide) and (not HasRightSide) then
    raise ELexerError.Create('Invalid number', FBufferReader.LineNumber, FBufferReader.ColumnNumber);

  ExponentSign := 1;
  // Next check for scientific notation
  if (FCurrentChar = 'e') or (FCurrentChar = 'E') then
  begin
    // Then it's a float. Start collecting exponent part
    if Result.Kind = TTokenKind.IntegerLiteral then
    begin
      Result.Kind := TTokenKind.FloatLiteral;
      Result.FloatValue := Result.IntegerValue;
    end;
    FCurrentChar := FBufferReader.ReadChar;
    if (FCurrentChar = '-') or (FCurrentChar = '+') then
    begin
      if FCurrentChar = '-' then
        ExponentSign := -1;
      FCurrentChar := FBufferReader.ReadChar;
    end;

    // accumulate exponent, check that first ch is a digit
    if not IsDigit (FCurrentChar) then
      raise ELexerError.Create ('Number expected in exponent', FBufferReader.LineNumber, FBufferReader.ColumnNumber);

    Evalue := 0;
    repeat

      SingleDigit := Ord(FCurrentChar) - Ord('0');
      if Evalue > (MAX_EXPONENT - SingleDigit) div 10 then
        raise ELexerError.Create ('Exponent overflow, maximum value for exponent is ' + IntToStr(MAX_EXPONENT), FBufferReader.LineNumber, FBufferReader.ColumnNumber);

      Evalue := 10*evalue + singleDigit;
      FCurrentChar := FBufferReader.ReadChar;

    until not IsDigit(FCurrentChar);

    evalue := evalue * ExponentSign;
    if Result.Kind = TTokenKind.IntegerLiteral then
      Result.FloatValue := Result.IntegerValue * IntPower(10, Evalue)
    else
      Result.FloatValue := Result.FloatValue * Power(10.0, Evalue);

  end;

end;

function TTokenScanner.GetSpecial: TToken;
begin
  case FCurrentChar of
    '&':
    begin
      if FBufferReader.ReadNextChar = '&' then
      begin
        Result.Kind := TTokenKind.LogicalAnd;
        FCurrentChar := FBufferReader.ReadChar;
      end
      else
        Result.Kind := TTokenKind.BinaryAnd;
    end;
    '|':
    begin
      if FBufferReader.ReadNextChar = '|' then
      begin
        Result.Kind := TTokenKind.LogicalOr;
        FCurrentChar := FBufferReader.ReadChar;
      end
      else
        Result.Kind := TTokenKind.BinaryOr;
    end;

    '>':
    begin
      if FBufferReader.ReadNextChar = '=' then
      begin
        Result.Kind := TTokenKind.MoreThenOrEqual;
        FCurrentChar := FBufferReader.ReadChar;
      end
      else
        Result.Kind := TTokenKind.MoreThen;
    end;
    '<':
    begin
      if FBufferReader.ReadNextChar = '=' then
      begin
        Result.Kind := TTokenKind.LessThenOrEqual;
        FCurrentChar := FBufferReader.ReadChar;
      end
      else
        Result.Kind := TTokenKind.LessThen;
    end;
    '!':
    begin
      if FBufferReader.ReadNextChar = '=' then
      begin
        Result.Kind := TTokenKind.NotEqual;
        FCurrentChar := FBufferReader.ReadChar;
      end
      else
        Result.Kind := TTokenKind.IdentifierNot;
    end;
    '+': Result.Kind := TTokenKind.Plus;
    '-': Result.Kind := TTokenKind.Minus;
    '*': Result.Kind := TTokenKind.Mult;
    '/': Result.Kind := TTokenKind.Divide;
    '^': Result.Kind := TTokenKind.Power;
    '=':
    begin
      if FBufferReader.ReadNextChar = '=' then
      begin
        Result.Kind := TTokenKind.Equivalence;
        FCurrentChar := FBufferReader.ReadChar;
      end
      else
        Result.Kind := TTokenKind.Assignment;
    end;
    ';': Result.Kind := TTokenKind.Semicolon;
    '(': Result.Kind := TTokenKind.LeftParenthesis;
    ')': Result.Kind := TTokenKind.RightParenthesis;
    '[': Result.Kind := TTokenKind.LeftSquareBracket;
    ']': Result.Kind := TTokenKind.RightSquareBracket;
    '{': Result.Kind := TTokenKind.LeftCurleyBracket;
    '}': Result.Kind := TTokenKind.RightCurleyBracket;
    ',': Result.Kind := TTokenKind.Comma;
    '#': Result.Kind := TTokenKind.Hash;
    else
      raise ELexerError.Create(Format('Invalid symbol "%s"', [FCurrentChar]), FBufferReader.LineNumber, FBufferReader.ColumnNumber);
  end;
  FCurrentChar := FBufferReader.ReadChar;
end;

function TTokenScanner.GetMultiLineString: TToken;
begin
  Result.Kind := TTokenKind.StringLiteral;
  Result.StringValue := '';

  // Remove the two quotes from buffer
  FBufferReader.ReadChar;
  FBufferReader.ReadChar;
  FCurrentChar := FBufferReader.ReadChar;
  while not IsEof(FCurrentChar) do
  begin
    if (FCurrentChar = '"') and (FBufferReader.ReadNextChar(1) = '"') and (FBufferReader.ReadNextChar(2) = '"') then
    begin
      // Remove the two quotes from buffer
      FBufferReader.ReadChar;
      FBufferReader.ReadChar;
      FCurrentChar := FBufferReader.ReadChar;
      Exit;
    end;

    Result.StringValue := Result.StringValue + FCurrentChar;
    FCurrentChar := FBufferReader.ReadChar;
  end;

end;

function TTokenScanner.GetString: TToken;
begin
  Result.Kind := TTokenKind.StringLiteral;
  Result.StringValue := '';

  FCurrentChar := FBufferReader.ReadChar;
  while not IsEof(FCurrentChar) do
  begin
    if FCurrentChar = '"' then
    begin
      FCurrentChar := FBufferReader.ReadChar;
      Exit;
    end;
    if FCurrentChar = '\' then
    begin
      FCurrentChar := FBufferReader.ReadChar;
      case FCurrentChar of
        '\': Result.StringValue := Result.StringValue + '\';
        'n': Result.StringValue := Result.StringValue + sLineBreak;
        'r': Result.StringValue := Result.StringValue + TBufferReader.CR;
        't': Result.StringValue := Result.StringValue + TBufferReader.TAB;
        '"': Result.StringValue := Result.StringValue + '"';
        '''': Result.StringValue := Result.StringValue + '''';
      end;
    end
    else
      Result.StringValue := Result.StringValue + FCurrentChar;
    FCurrentChar := FBufferReader.ReadChar;
  end;
  raise ELexerError.Create('Unterminated string', FBufferReader.LineNumber, FBufferReader.ColumnNumber);
end;

function TTokenScanner.GetWord: TToken;
var
  LTokenKind: TTokenKind;
begin
  Result.Kind := TTokenKind.Identifier;
  Result.StringValue := '';

  while IsDigit(FCurrentChar) or IsLetter(FCurrentChar) do
  begin
    Result.StringValue := Result.StringValue + FCurrentChar;
    FCurrentChar := FBufferReader.ReadChar;
  end;

  if IsKeyword(Result.StringValue, LTokenKind) then
    Result.Kind := LTokenKind;
end;

procedure TTokenScanner.InitKeywords;
begin
//  FKeywords.Add('if', TTokenKind.KeywordIf);
//  FKeywords.Add('end', TTokenKind.KeywordEnd);
//  FKeywords.Add('println', TTokenKind.KeywordPrintln);
//  FKeywords.Add('then', TTokenKind.KeywordThen);
//  FKeywords.Add('until', TTokenKind.KeywordUntil);
//  FKeywords.Add('else', TTokenKind.KeywordElse);
//  FKeywords.Add('while', TTokenKind.KeywordWhile);
//  FKeywords.Add('for', TTokenKind.KeywordFor);
//  FKeywords.Add('repeat', TTokenKind.KeywordRepeat);
//  FKeywords.Add('return', TTokenKind.KeywordReturn);
//  FKeywords.Add('break', TTokenKind.KeywordBreak);
//  FKeywords.Add('function', TTokenKind.KeywordFunction);
//  FKeywords.Add('do', TTokenKind.KeywordDo);
//  FKeywords.Add('to', TTokenKind.KeywordTo);
//  FKeywords.Add('downto', TTokenKind.KeywordDownto);
//  FKeywords.Add('ref', TTokenKind.KeywordRef);
//  FKeywords.Add('not', TTokenKind.KeywordNot);
  FKeywords.Add('true', TTokenKind.KeywordTrue);
  FKeywords.Add('false', TTokenKind.KeywordFalse);
  FKeywords.Add('null', TTokenKind.KeywordNull);
end;

function TTokenScanner.IsKeyword(const AIdentifier: string; out ATokenKind: TTokenKind): Boolean;
begin
  Result := FKeywords.TryGetValue(AIdentifier, ATokenKind);
end;

constructor TTokenScanner.CreateFromFile(const AFileName: TFileName; AEncoding: TEncoding);
begin
  CreateFromStream(TBufferedFileStream.Create(AFileName, fmOpenRead), AEncoding, True);
end;

constructor TTokenScanner.CreateFromStream(AStream: TStream; AEncoding: TEncoding; AOwnStream: Boolean);
begin
  FBufferReader := TBufferReader.Create(AStream, AEncoding, AOwnStream);
  FKeywords := TDictionary<string, TTokenKind>.Create;
  InitKeywords;

  FCurrentChar := FBufferReader.ReadChar;
end;

constructor TTokenScanner.CreateFromString(const AValue: string);
begin
  CreateFromStream(TStringStream.Create(AValue, TEncoding.UTF8), TEncoding.UTF8, True);
end;

procedure TTokenScanner.SkipBlankAndComment;
begin
  while IsBlank(FCurrentChar) or (FCurrentChar = '/') or (FCurrentChar = '#') do
  begin
    if IsBlank(FCurrentChar) then
    begin
      FCurrentChar := FBufferReader.ReadChar;
    end
    else if FCurrentChar = '#' then
    begin
      FCurrentChar := FBufferReader.ReadChar;
      SkipSingleLineComment;
    end
    else if FCurrentChar = '/' then
    begin
      if FBufferReader.ReadNextChar = '/' then
      begin
        FCurrentChar := FBufferReader.ReadChar;
        SkipSingleLineComment;
      end
      else if FBufferReader.ReadNextChar = '*' then
      begin
        FCurrentChar := FBufferReader.ReadChar;
        SkipMultiLineComment;
      end
      else
        Break;
    end;
  end;
end;

procedure TTokenScanner.SkipMultiLineComment;
begin
  FCurrentChar := FBufferReader.ReadChar;
  while True do
  begin
    while (FCurrentChar <> '*') and (not IsEof(FCurrentChar)) do
      FCurrentChar := FBufferReader.ReadChar;
    if IsEof(FCurrentChar) then
      raise ELexerError.Create('Unterminated comment', FBufferReader.LineNumber, FBufferReader.ColumnNumber);

    FCurrentChar := FBufferReader.ReadChar;
    if FCurrentChar = '/' then
    begin
      FCurrentChar := FBufferReader.ReadChar;
      Exit;
    end;
  end;
  raise ELexerError.Create('Unterminated comment', FBufferReader.LineNumber, FBufferReader.ColumnNumber);
end;

procedure TTokenScanner.SkipSingleLineComment;
begin
  while (FCurrentChar <> TBufferReader.LF) and (not IsEof(FCurrentChar)) do
    FCurrentChar := FBufferReader.ReadChar;
  if not IsEof(FCurrentChar) then
    FCurrentChar := FBufferReader.ReadChar;
end;

function TTokenScanner.NextToken: TToken;
begin
  if not Assigned(FBufferReader) then
    raise ELexerError.Create('Parsing not yet started', FBufferReader.LineNumber, FBufferReader.ColumnNumber);

  SkipBlankAndComment;

  Result.LineNumber := FBufferReader.LineNumber;
  Result.ColumnNumber := FBufferReader.ColumnNumber;

  if IsLetter(FCurrentChar) then
  begin
    Result := GetWord;
    Exit;
  end;

  if IsDigit(FCurrentChar) then
  begin
    Result := GetNumber;
    Exit;
  end;

  if FCurrentChar = '"' then
  begin
    if (FBufferReader.ReadNextChar = '"') and (FBufferReader.ReadNextChar(2) = '"') then
      Result := GetMultiLineString
    else
      Result := GetString;
    Exit;
  end;

  if IsEof(FCurrentChar) then
  begin
    Result.Kind := TTokenKind.EndOfStream;
    Exit;
  end;

  Result := GetSpecial;
end;

{ TToken }

function TToken.ToString: string;
begin
  Result := '(' + KindToString(Kind) + ')';
  case KindToType(Kind) of
    varInteger: Result := Result + IntToStr(IntegerValue);
    varUString: Result := Result + StringValue;
    varDouble: Result := Result + FloatToStr(FloatValue);
  end;
end;

{ ELexerError }

constructor ELexerError.Create(const AMessage: string; ALine, ACol: Integer);
begin
  FLine := ALine;
  FCol := ACol;
  inherited CreateFmt(AMessage + ' at line %d', [ALine]);
end;

end.
