unit DBBuilder.Core.SyntaxAnalysis;

interface

uses
  System.Classes, System.SysUtils, System.Rtti, Generics.Collections,
  DBBuilder.Core.Lexer;

type
  ESyntaxError = class(Exception)
  private
    FCol: Integer;
    FLine: Integer;
  public
    property Line: Integer read FLine;
    property Col: Integer read FCol;
    constructor Create(const AMessage: string; ALine, ACol: Integer); overload;
    constructor Create(const AMessage: string; AToken: TToken); overload;
    constructor CreateFmt(const AMessage: string; AArgs: array of const; AToken: TToken); overload;
  end;

  TSyntaxAnalysis = class(TObject)
  private
    FTokenQueue: TQueue<TToken>;
  protected
    FOwnScanner: Boolean;
    FScanner: TTokenScanner;
    FToken: TToken;
    procedure Expect(ATokenKind: TTokenKind; MoveNext: Boolean = True); overload;
    // Case sensitive
    procedure Expect(ATokenKind: TTokenKind; const AStringValue: string; MoveNext: Boolean = True); overload;
    // Case insensitive
    procedure ExpectText(ATokenKind: TTokenKind; const AStringValue: string; MoveNext: Boolean = True);
    procedure NextToken; virtual;
    function Lookahead: TToken; virtual;
  public
    constructor Create(AScanner: TTokenScanner; AOwnScanner: Boolean = False); virtual;
    destructor Destroy; override;

    function Execute: TValue; virtual; abstract;
  end;

implementation

{ ESyntaxError }

constructor ESyntaxError.Create(const AMessage: string; ALine, ACol: Integer);
begin
  FLine := ALine;
  FCol := ACol;
  inherited CreateFmt(AMessage + ' at line %d', [ALine]);
end;

constructor ESyntaxError.Create(const AMessage: string; AToken: TToken);
begin
  Create(AMessage, AToken.LineNumber, AToken.ColumnNumber);
end;

constructor ESyntaxError.CreateFmt(const AMessage: string;
  AArgs: array of const; AToken: TToken);
begin
  Create(Format(AMessage, AArgs), AToken.LineNumber, AToken.ColumnNumber);
end;

{ TSyntaxAnalysis }

constructor TSyntaxAnalysis.Create(AScanner: TTokenScanner; AOwnScanner: Boolean = False);
begin
  inherited Create;
  FScanner := AScanner;
  FOwnScanner := AOwnScanner;
  FTokenQueue := TQueue<TToken>.Create;
end;

procedure TSyntaxAnalysis.NextToken;
begin
  if FTokenQueue.Count > 0 then
  begin
    FToken := FTokenQueue.Dequeue;
    Exit;
  end;
  FToken := FScanner.NextToken;
end;

function TSyntaxAnalysis.Lookahead: TToken;
begin
  Result := FScanner.NextToken;
  FTokenQueue.Enqueue(Result);
end;

destructor TSyntaxAnalysis.Destroy;
begin
  FTokenQueue.Free;
  if FOwnScanner then
    FScanner.Free;
  inherited;
end;

procedure TSyntaxAnalysis.Expect(ATokenKind: TTokenKind;
  const AStringValue: string; MoveNext: Boolean);
begin
  if (ATokenKind <> FToken.Kind) or (AStringValue <> FToken.StringValue) then
    raise ESyntaxError.Create(
      Format('Expected [%s] with value [%s] but found [%s] with value [%s]', [
        KindToString(ATokenKind), AStringValue, KindToString(FToken.Kind), FToken.StringValue]),
      FToken.LineNumber,
      FToken.ColumnNumber
    );
  if MoveNext then
    NextToken;
end;

procedure TSyntaxAnalysis.ExpectText(ATokenKind: TTokenKind;
  const AStringValue: string; MoveNext: Boolean);
begin
  if (ATokenKind <> FToken.Kind) or (not SameText(AStringValue, FToken.StringValue)) then
    raise ESyntaxError.Create(
      Format('Expected [%s] with value [%s] but found [%s] with value [%s]', [
        KindToString(ATokenKind), AStringValue, KindToString(FToken.Kind), FToken.StringValue]),
      FToken.LineNumber,
      FToken.ColumnNumber
    );
  if MoveNext then
    NextToken;
end;

procedure TSyntaxAnalysis.Expect(ATokenKind: TTokenKind; MoveNext: Boolean);
begin
  if ATokenKind <> FToken.Kind then
    raise ESyntaxError.Create(
      Format('Expected [%s] but found [%s]', [KindToString(ATokenKind), KindToString(FToken.Kind)]),
      FToken.LineNumber,
      FToken.ColumnNumber
    );
  if MoveNext then
    NextToken;
end;

end.
