unit DBBuilder.DB.DatabaseManager;

interface

uses
  System.SysUtils, System.Classes, System.Hash, System.IOUtils,
  DBBuilder.DB.DatabaseEngine, Data.DB, DBBuilder.Core.Config, DBBuilder.DB.Core;

type
  TDatabaseManager = class(TObject)
  private
    function GetConnectionString: string;
  protected
    FScriptDataSet: TDataSet;
    FDatabaseEngine: IDatabaseEngine;
    FDatabase: TDatabaseInfo;
    procedure InitScriptTable; virtual; abstract;
    procedure CheckConnect;
    function GetName: string; virtual; abstract;

    property DatabaseEngine: IDatabaseEngine read FDatabaseEngine;
  public
    procedure LoadScriptTable;
    property Name: string read GetName;
    property ConnectionString: string read GetConnectionString;
    procedure AlterTriggerStatus(const ATriggerName: string; ATriggerStatus: TTriggerStatus);
    procedure AlterAllTriggerStatus(ATriggerStatus: TTriggerStatus);
    procedure SetSequence(const ASequenceName: string; AValue: Int64);
    function QuoteValue(AValue: Variant; AFieldType: TFieldType): string;
    function CheckForScript(const AScriptName: string): Boolean;
    function CreateDataSet(const ASQL: string): TDataSet;
    procedure StoreAction(const ActionName: string);
    procedure ExecuteScript(const AScript: string);
    procedure ExecuteStatement(const AScript: string; AParams: TDBParams);
    procedure ClearCache;
    procedure Commit;
    procedure Rollback;

    constructor Create(ADatabase: TDatabaseInfo); reintroduce;
    destructor Destroy; override;
  end;

  TTargetDatabaseManager = class(TDatabaseManager)
  protected
    procedure InitScriptTable; override;
    function GetName: string; override;
  end;

  TSourceDatabaseManager = class(TDatabaseManager)
  protected
    procedure InitScriptTable; override;
    function GetName: string; override;
  end;

function TargetDatabaseManager: TDatabaseManager;
function SourceDatabaseManager: TDatabaseManager;

implementation

uses
  Dynamo.Core.ServiceLocator;

var
  _TargetDatabaseManager: TDatabaseManager;
  _SourceDatabaseManager: TDatabaseManager;

const
  StmtInsertScript =
    ' insert into SCRIPTS (ID, SCRIPT_NAME, FILE_HASH) ' + sLineBreak +
    ' values (GEN_ID(GEN_SCRIPT, 1), %s, %s) ';


function TargetDatabaseManager: TDatabaseManager;
begin
  if not Assigned(_TargetDatabaseManager) then
    _TargetDatabaseManager := TTargetDatabaseManager.Create(Config.TargetDatabase);
  Result := _TargetDatabaseManager;
end;

function SourceDatabaseManager: TDatabaseManager;
begin
  if not Assigned(_SourceDatabaseManager) then
    _SourceDatabaseManager := TSourceDatabaseManager.Create(Config.SourceDatabase);
  Result := _SourceDatabaseManager;
end;

{ TDatabaseManager }

procedure TDatabaseManager.AlterAllTriggerStatus(
  ATriggerStatus: TTriggerStatus);
var
  LTriggers: TArray<string>;
  LTriggerName: string;
begin
  LTriggers := DatabaseEngine.GetAllTriggers;
  for LTriggerName in LTriggers do
    DatabaseEngine.AlterTriggerStatus(LTriggerName, ATriggerStatus);
end;

procedure TDatabaseManager.AlterTriggerStatus(const ATriggerName: string;
  ATriggerStatus: TTriggerStatus);
begin
  CheckConnect;
  DatabaseEngine.AlterTriggerStatus(ATriggerName, ATriggerStatus);
end;

procedure TDatabaseManager.CheckConnect;
begin
  if not DatabaseEngine.Connected then
  begin
    DatabaseEngine.Connect(FDatabase);
  end;
end;

function TDatabaseManager.CheckForScript(const AScriptName: string): Boolean;
begin
  CheckConnect;
  InitScriptTable;
  if not Assigned(FScriptDataSet) then
    Exit(False);

  Result := FScriptDataSet.Locate('SCRIPT_NAME', AScriptName, []);
end;

procedure TDatabaseManager.ClearCache;
begin
  FreeAndNil(FScriptDataSet);
end;

procedure TDatabaseManager.Commit;
begin
  DatabaseEngine.Commit;
end;

constructor TDatabaseManager.Create(ADatabase: TDatabaseInfo);
begin
  inherited Create;
  FDatabaseEngine := ServiceLocator.GetClass<IDatabaseEngine>(ADatabase.Engine);
  FDatabase := ADatabase;
end;

function TDatabaseManager.CreateDataSet(const ASQL: string): TDataSet;
begin
  CheckConnect;
  Result := DatabaseEngine.CreateDataSet(ASQL);
end;

destructor TDatabaseManager.Destroy;
begin
  FScriptDataSet.Free;
  inherited;
end;

procedure TDatabaseManager.ExecuteStatement(const AScript: string;
  AParams: TDBParams);
begin
  CheckConnect;
  DatabaseEngine.ExecuteStatement(AScript, AParams);
end;

procedure TDatabaseManager.ExecuteScript(const AScript: string);
begin
  CheckConnect;
  DatabaseEngine.ExecuteScript(AScript);
  InitScriptTable;
end;

function TDatabaseManager.GetConnectionString: string;
begin
  Result := FDatabase.ConnectionString;
end;

procedure TDatabaseManager.LoadScriptTable;
begin
  InitScriptTable;
end;

function TDatabaseManager.QuoteValue(AValue: Variant;
  AFieldType: TFieldType): string;
begin
  Result := DatabaseEngine.QuoteValue(AValue, AFieldType);
end;

procedure TDatabaseManager.Rollback;
begin
  DatabaseEngine.Rollback;
end;

procedure TDatabaseManager.SetSequence(const ASequenceName: string;
  AValue: Int64);
begin
  CheckConnect;
  DatabaseEngine.SetSequence(ASequenceName, AValue);
end;

procedure TDatabaseManager.StoreAction(const ActionName: string);
var
  LFileHash: string;
begin
  CheckConnect;
  if not CheckForScript(ActionName) then
  begin
    LFileHash := THashMD5.GetHashStringFromFile(TPath.Combine(Config.ScriptPath, ActionName));
    DatabaseEngine.ExecuteScript(Format(StmtInsertScript, [QuotedStr(ActionName), QuotedStr(LFileHash)]));
    DatabaseEngine.Commit;
  end;
end;

{ TTargetDatabaseManager }

function TTargetDatabaseManager.GetName: string;
begin
  Result := 'TARGET';
end;

procedure TTargetDatabaseManager.InitScriptTable;
begin
  if DatabaseEngine.CheckTable('SCRIPTS') and (not Assigned(FScriptDataSet)) then
  begin
    FScriptDataSet := DatabaseEngine.CreateDataSet('SELECT * FROM SCRIPTS');
  end;
end;

{ TSourceDatabaseManager }

function TSourceDatabaseManager.GetName: string;
begin
  Result := 'SOURCE';
end;

procedure TSourceDatabaseManager.InitScriptTable;
begin
end;

initialization

  _TargetDatabaseManager := nil;
  _SourceDatabaseManager := nil;

finalization

  _TargetDatabaseManager.Free;
  _SourceDatabaseManager.Free;

end.
