unit DBBuilder.DB.FDDatabaseEngine;

interface

uses
  System.SysUtils, System.Classes, System.DateUtils, System.TypInfo,
  FireDAC.Stan.Intf, FireDAC.Stan.Option,
  FireDAC.Stan.Error, FireDAC.UI.Intf,
  FireDAC.Stan.Pool, FireDAC.Stan.Async, FireDAC.VCLUI.Wait,
  Data.DB, FireDAC.Comp.Client, DBBuilder.DB.DatabaseEngine, Dynamo.Core.Rtti,
  Dynamo.Core.ServiceLocator, FireDAC.Comp.ScriptCommands, FireDAC.Stan.Util,
  FireDAC.Comp.Script, FireDAC.Stan.Param, FireDAC.DatS, FireDAC.DApt.Intf,
  FireDAC.DApt, FireDAC.Comp.DataSet, DBBuilder.Core.Config, System.Hash,
  DBBuilder.DB.Core;

type
  TFDDatabaseEngine = class(TInterfacedObject, IDatabaseEngine)
  protected
    FDConnection: TFDConnection;
    FDScript: TFDScript;
    FDQuery: TFDQuery;
  public
    { IDatabaseDriver }
    function GetAllTriggers: System.TArray<System.string>; virtual; abstract;
    procedure AlterTriggerStatus(const ATriggerName: string; ATriggerStatus: TTriggerStatus); virtual; abstract;
    procedure SetSequence(const ASequenceName: string; AValue: Int64); virtual; abstract;
    function DateTimeToDBStr(ADateTime: TDateTime): string; virtual; abstract;
    function DateToDBStr(ADateTime: TDateTime): string; virtual; abstract;
    function QuoteValue(AValue: Variant; AFieldType: TFieldType): string; virtual; abstract;
    function CheckTable(const ATableName: string): Boolean; virtual; abstract;
    procedure ExecuteScript(const AStatement: string);
    procedure ExecuteStatement(const AStatement: string; AParams: TDBParams);
    function CreateDataSet(const ASQL: string): TDataSet;
    function GetConnect: Boolean;
    procedure Connect(ADatabase: TDatabaseInfo); virtual; abstract;
    procedure Commit;
    procedure Rollback;
    procedure Disconnect;

    constructor Create; virtual;
    destructor Destroy; override;
  end;

implementation

{ TFireDacDatabaseEngine }

procedure TFDDatabaseEngine.Commit;
begin
  FDConnection.Commit;
end;

constructor TFDDatabaseEngine.Create;
begin
  inherited;
  FDConnection := TFDConnection.Create(nil);
  FDScript := TFDScript.Create(nil);
  FDScript.Connection := FDConnection;

  FDQuery := TFDQuery.Create(nil);
  FDQuery.Connection := FDConnection;
end;

function TFDDatabaseEngine.CreateDataSet(const ASQL: string): TDataSet;
var
  LQuery: TFDQuery;
begin
  LQuery := TFDQuery.Create(nil);
  try
    LQuery.Connection := FDConnection;
    LQuery.SQL.Text := ASQL;
    LQuery.Open;
  except
    LQuery.Free;
    raise;
  end;
  Result := LQuery;
end;

destructor TFDDatabaseEngine.Destroy;
begin
  Disconnect;
  FDScript.Free;
  FDQuery.Free;
  FDConnection.Free;
  inherited;
end;

procedure TFDDatabaseEngine.Disconnect;
begin
  FDConnection.Connected := False;
end;

procedure TFDDatabaseEngine.ExecuteScript(const AStatement: string);
var
  LStatement: TStrings;
begin
  FDScript.Params.Clear;
  LStatement := TStringList.Create;
  try
    LStatement.Text := AStatement;
    FDScript.ScriptOptions.BreakOnError := True;
    FDScript.ExecuteScript(LStatement);
  finally
    LStatement.Free;
  end;
end;

procedure TFDDatabaseEngine.ExecuteStatement(const AStatement: string;
  AParams: TDBParams);
var
  LQuery: TFDQuery;
  LParam: TDBParam;
  LFDParam: TFDParam;
begin
  LQuery := TFDQuery.Create(nil);
  try
    LQuery.Connection := FDConnection;
    LQuery.SQL.Text := AStatement;

    if Assigned(AParams) then
    begin
      for LParam in AParams.ToArray do
      begin
        LFDParam := LQuery.ParamByName(LParam.Name);
        LFDParam.Value := LParam.Value;
      end;
    end;

    LQuery.ExecSQL;
  finally
    LQuery.Free;
  end;
end;

function TFDDatabaseEngine.GetConnect: Boolean;
begin
  Result := FDConnection.Connected;
end;

procedure TFDDatabaseEngine.Rollback;
begin
  FDConnection.Rollback;
end;

end.
