unit DBBuilder.DB.FDFirebirdDatabaseEngine;

interface

uses
  System.SysUtils, System.Classes, System.DateUtils, System.TypInfo,
  System.Variants, FireDAC.Phys.IBWrapper,
  FireDAC.Stan.Intf, FireDAC.Stan.Option,
  FireDAC.Stan.Error, FireDAC.UI.Intf, FireDAC.Phys.Intf, FireDAC.Stan.Def,
  FireDAC.Stan.Pool, FireDAC.Stan.Async, FireDAC.Phys, FireDAC.VCLUI.Wait,
  Data.DB, FireDAC.Comp.Client, FireDAC.Phys.FBDef, FireDAC.Phys.IBBase,
  FireDAC.Phys.FB, DBBuilder.DB.DatabaseEngine, Dynamo.Core.Rtti,
  Dynamo.Core.ServiceLocator, FireDAC.Comp.ScriptCommands, FireDAC.Stan.Util,
  FireDAC.Comp.Script, FireDAC.Stan.Param, FireDAC.DatS, FireDAC.DApt.Intf,
  FireDAC.DApt, FireDAC.Comp.DataSet, DBBuilder.Core.Config, System.Hash,
  DBBuilder.DB.FDDatabaseEngine, DBBuilder.DB.Core;

type
  [Alias('FDFirebird')]
  TFDFirebirdDatabaseEngine = class(TFDDatabaseEngine)
  private
    FBDriverLink: TFDPhysFBDriverLink;
  public
    function GetAllTriggers: System.TArray<System.string>; override;
    procedure AlterTriggerStatus(const ATriggerName: string; ATriggerStatus: TTriggerStatus); override;
    procedure SetSequence(const ASequenceName: string; AValue: Int64); override;
    function DateTimeToDBStr(ADateTime: TDateTime): string; override;
    function DateToDBStr(ADateTime: TDateTime): string; override;
    procedure Connect(ADatabase: TDatabaseInfo); override;
    function CheckTable(const ATableName: string): Boolean; override;
    function QuoteValue(AValue: Variant; AFieldType: TFieldType): string; override;

    constructor Create; override;
    destructor Destroy; override;
  end;

implementation

uses
  DBBuilder.Core.Logging;

var
  FBFormatSettings: TFormatSettings;

procedure InitFormatSettings;
begin
  FBFormatSettings := TFormatSettings.Invariant;
  FBFormatSettings.ShortDateFormat := 'yyyy-mm-dd';
  FBFormatSettings.DateSeparator := '-';
  FBFormatSettings.DecimalSeparator := '.';
end;


procedure TFDFirebirdDatabaseEngine.AlterTriggerStatus(
  const ATriggerName: string; ATriggerStatus: TTriggerStatus);
const
  cTriggerStatusStr: array [TTriggerStatus] of string = ('INACTIVE', 'ACTIVE');
begin
  inherited;
  TLogger.Log(llDebug, Format('ALTER TRIGGER %s %s', [ATriggerName, cTriggerStatusStr[ATriggerStatus]]));

  ExecuteScript(Format('ALTER TRIGGER %s %s', [ATriggerName, cTriggerStatusStr[ATriggerStatus]]));
end;

function TFDFirebirdDatabaseEngine.CheckTable(
  const ATableName: string): Boolean;
begin
  FDQuery.Close;
  FDQuery.SQL.Text := 'SELECT 1 FROM RDB$RELATIONS WHERE RDB$RELATION_NAME = ' + QuotedStr(ATableName);
  FDQuery.Open;
  Result := not FDQuery.IsEmpty;
  FDQuery.Close;
end;

function ParseCharSet(const s: string): TIBCharacterSet;
begin
  if CompareText(s, '') = 0 then
    Result := csNONE
  else if CompareText(s, 'NONE') = 0 then
    Result := csNONE
  else if CompareText(s, 'ASCII') = 0 then
    Result := csASCII
  else if CompareText(s, 'UNICODE_FSS') = 0 then
    Result := csUNICODE_FSS
  else if CompareText(s, 'WIN1250') = 0 then
    Result := csWIN1250
  else if CompareText(s, 'WIN1251') = 0 then
    Result := csWIN1251
  else if CompareText(s, 'WIN1252') = 0 then
    Result := csWIN1252
  else if CompareText(s, 'ISO8859_1') = 0 then
    Result := csISO8859_1
  else if CompareText(s, 'UTF8') = 0 then
    Result := csUTF8
  else
    raise Exception.CreateFmt('Unknown charset "%s"', [s]);
end;

procedure TFDFirebirdDatabaseEngine.Connect(ADatabase: TDatabaseInfo);
var
  LParams: TFDPhysFBConnectionDefParams;
  //DriverId: string;
begin
  if not FDConnection.Connected then
  begin
    //DriverId := THashMD5.GetHashString(ADatabase.Server + IntToStr(ADatabase.Port) + ADatabase.Database);
    //FPhysFBDriverLink.VendorHome := 'C:\Firebird\Firebird-4.0.0.1436-0';
    //FBDriverLink.VendorHome := '';
    //FBDriverLink.DriverID := DriverId;
    FBDriverLink.VendorHome := ExtractFilePath(ADatabase.ClientLib);
    FBDriverLink.VendorLib := ExtractFileName(ADatabase.ClientLib);

    FDConnection.DriverName := 'FB';
    LParams := FDConnection.Params as TFDPhysFBConnectionDefParams;
    LParams.Server := ADatabase.Server;
    if ADatabase.Port > 0 then
      LParams.Port := ADatabase.Port;
    if ADatabase.Charset <> '' then
      LParams.CharacterSet := ParseCharSet(ADatabase.Charset);
    LParams.Database := ADatabase.Database;
    LParams.UserName := ADatabase.UserName;
    LParams.Password := ADatabase.Password;

    FDConnection.LoginPrompt := False;
    FDConnection.Connected := True;
  end;
end;

constructor TFDFirebirdDatabaseEngine.Create;
begin
  inherited;
  FBDriverLink := TFDPhysFBDriverLink.Create(nil);
end;

function TFDFirebirdDatabaseEngine.DateTimeToDBStr(ADateTime: TDateTime): string;
begin
  Result := '''' + FormatDateTime('yyyy-mm-dd hh:nn:ss', ADateTime) + '''';
end;

function TFDFirebirdDatabaseEngine.DateToDBStr(ADateTime: TDateTime): string;
begin
  Result := '''' + FormatDateTime('yyyy-mm-dd', ADateTime) + '''';
end;

destructor TFDFirebirdDatabaseEngine.Destroy;
begin
  FBDriverLink.Free;
  inherited;
end;

function TFDFirebirdDatabaseEngine.GetAllTriggers: System.TArray<System.string>;
const
  stmtTrigger =
    ' SELECT * FROM RDB$TRIGGERS ' +
    ' WHERE RDB$TRIGGER_NAME NOT LIKE ''RDB$%'' ';
var
  LDataSet: TDataSet;
begin
  Result := [];
  LDataSet := CreateDataSet(stmtTrigger);
  try
    LDataSet.First;
    while not LDataSet.Eof do
    begin
      Result := Result + [LDataSet.FieldByName('RDB$TRIGGER_NAME').AsString];
      LDataSet.Next;
    end;
  finally
    LDataSet.Free;
  end;
end;

function TFDFirebirdDatabaseEngine.QuoteValue(AValue: Variant;
  AFieldType: TFieldType): string;
begin
  if VarIsNull(AValue) then
    Exit('NULL');

  case AFieldType of
    ftString, ftFixedChar, ftWideString:
      Result := QuotedStr(AValue);
    ftAutoInc, ftLongWord, ftShortint, ftByte, ftSmallint, ftInteger, ftWord, ftLargeint:
      Result := IntToStr(AValue);
    ftFloat, ftCurrency, ftBCD, ftExtended, ftSingle:
      Result := FloatToStr(AValue, FBFormatSettings);
    ftDate:
      Result := DateToDBStr(AValue);
    ftDateTime, ftTimeStamp:
      Result := DateTimeToDBStr(AValue);
    else
      raise EDatabaseEngineError.CreateFmt('Tipo "%s" non supportato', [GetEnumName(TypeInfo(TFieldType), Integer(AFieldType))]);
  end;
end;

procedure TFDFirebirdDatabaseEngine.SetSequence(const ASequenceName: string;
  AValue: Int64);
begin
  ExecuteScript('SET GENERATOR ' + ASequenceName + ' TO ' + IntToStr(AValue));
end;

initialization

  ServiceLocator.RegisterClass(TFDFirebirdDatabaseEngine);
  InitFormatSettings;

end.
