unit DBBuilder.DB.FDPostgresDatabaseEngine;

interface

uses
  System.SysUtils, System.Classes, System.DateUtils, System.TypInfo,
  System.Variants,
  FireDAC.Stan.Intf, FireDAC.Stan.Option,
  FireDAC.Stan.Error, FireDAC.UI.Intf, FireDAC.Phys.Intf, FireDAC.Stan.Def,
  FireDAC.Stan.Pool, FireDAC.Stan.Async, FireDAC.Phys, FireDAC.VCLUI.Wait,
  Data.DB, FireDAC.Comp.Client,
  DBBuilder.DB.DatabaseEngine, Dynamo.Core.Rtti,
  Dynamo.Core.ServiceLocator, FireDAC.Comp.ScriptCommands, FireDAC.Stan.Util,
  FireDAC.Comp.Script, FireDAC.Stan.Param, FireDAC.DatS, FireDAC.DApt.Intf,
  FireDAC.DApt, FireDAC.Comp.DataSet, DBBuilder.Core.Config, System.Hash,
  FireDAC.Phys.PGDef, FireDAC.Phys.PG, DBBuilder.DB.FDDatabaseEngine,
  DBBuilder.DB.Core;

type
  [Alias('FDPostgres')]
  TFDPostgresDatabaseEngine = class(TFDDatabaseEngine)
  private
    PGDriverLink: TFDPhysPgDriverLink;
  public
    function GetAllTriggers: System.TArray<System.string>; override;
    procedure AlterTriggerStatus(const ATriggerName: string; ATriggerStatus: TTriggerStatus); override;
    procedure SetSequence(const ASequenceName: string; AValue: Int64); override;
    function DateTimeToDBStr(ADateTime: TDateTime): string; override;
    function DateToDBStr(ADateTime: TDateTime): string; override;
    procedure Connect(ADatabase: TDatabaseInfo); override;
    function CheckTable(const ATableName: string): Boolean; override;
    function QuoteValue(AValue: Variant; AFieldType: TFieldType): string; override;

    constructor Create; override;
    destructor Destroy; override;
  end;

implementation

var
  PGFormatSettings: TFormatSettings;

procedure InitFormatSettings;
begin
  PGFormatSettings := TFormatSettings.Invariant;
  PGFormatSettings.ShortDateFormat := 'yyyy-mm-dd';
  PGFormatSettings.DateSeparator := '-';
  PGFormatSettings.DecimalSeparator := '.';
end;

{ TFDPostgresDatabaseEngine }

procedure TFDPostgresDatabaseEngine.AlterTriggerStatus(
  const ATriggerName: string; ATriggerStatus: TTriggerStatus);
begin
  inherited;
  raise EDatabaseEngineError.Create('"AlterTriggerStatus" non implementato');
end;

function TFDPostgresDatabaseEngine.CheckTable(
  const ATableName: string): Boolean;
begin
  raise Exception.CreateFmt('"CheckTable" non supportato per la classe "%s"', [ClassName]);
end;

procedure TFDPostgresDatabaseEngine.Connect(ADatabase: TDatabaseInfo);
const
  cNameConnDef = 'PG_CONNECTION1';
 var
  oDef: IFDStanConnectionDef;
  oParams: TFDPhysPGConnectionDefParams;
  DriverId: string;
begin
  if not FDConnection.Connected then
  begin
    DriverId := THashMD5.GetHashString(ADatabase.Server + IntToStr(ADatabase.Port) + ADatabase.Database);
    PGDriverLink.VendorLib := ADatabase.ClientLib;
    PGDriverLink.DriverID := DriverId;

    oDef := FDManager.ConnectionDefs.FindConnectionDef(cNameConnDef);
    if not Assigned(oDef) then
    begin
      oDef := FDManager.ConnectionDefs.AddConnectionDef;
      oDef.Name := cNameConnDef;
    end;

    oParams := TFDPhysPGConnectionDefParams(oDef.Params);
    oParams.DriverID := DriverId;
    oParams.Database := ADatabase.Database;
    oParams.UserName := ADatabase.UserName;
    oParams.Password := ADatabase.Password;
    oParams.Server := ADatabase.Server;
    oDef.Apply;

    FDConnection.ConnectionDefName := cNameConnDef;
    FDConnection.Connected := True;
  end;
end;

constructor TFDPostgresDatabaseEngine.Create;
begin
  inherited;
  PGDriverLink := TFDPhysPgDriverLink.Create(nil);
end;

function TFDPostgresDatabaseEngine.DateTimeToDBStr(ADateTime: TDateTime): string;
begin
  Result := '''' + DateToISO8601(ADateTime, True) + '''';
end;

function TFDPostgresDatabaseEngine.DateToDBStr(ADateTime: TDateTime): string;
begin
  Result := '''' + FormatDateTime('yyyy-mm-dd', ADateTime) + '''';
end;

destructor TFDPostgresDatabaseEngine.Destroy;
begin
  PGDriverLink.Free;
  inherited;
end;

function TFDPostgresDatabaseEngine.GetAllTriggers: System.TArray<System.string>;
begin
  raise EDatabaseEngineError.Create('"GetAllTriggers" non implementato');
end;

function TFDPostgresDatabaseEngine.QuoteValue(AValue: Variant;
  AFieldType: TFieldType): string;
begin
  if VarIsNull(AValue) then
    Exit('NULL');

  case AFieldType of
    ftString, ftFixedChar, ftWideString:
      Result := QuotedStr(AValue);
    ftAutoInc, ftLongWord, ftShortint, ftByte, ftSmallint, ftInteger, ftWord, ftLargeint:
      Result := IntToStr(AValue);
    ftFloat, ftCurrency, ftBCD, ftExtended, ftSingle:
      Result := FloatToStr(AValue, PGFormatSettings);
    ftDate:
      Result := DateToDBStr(AValue);
    ftDateTime, ftTimeStamp:
      Result := DateTimeToDBStr(AValue);
    else
      raise EDatabaseEngineError.CreateFmt('Tipo "%s" non supportato', [GetEnumName(TypeInfo(TFieldType), Integer(AFieldType))]);
  end;
end;

procedure TFDPostgresDatabaseEngine.SetSequence(const ASequenceName: string;
  AValue: Int64);
begin
  raise EDatabaseEngineError.Create('"SetSequence" non implementato');
end;

initialization

  ServiceLocator.RegisterClass(TFDPostgresDatabaseEngine);
  InitFormatSettings;

end.
