unit DBBuilder.ISCScript.Interpreter;

interface

uses
  System.Classes, System.SysUtils, System.RTTI, System.Variants,
  DBBuilder.Core.SyntaxAnalysis,
  DBBuilder.Core.Logging, DBBuilder.Core.Lexer,
  DBBuilder.ISCScript.Processor;

type
  TRunner = class(TSyntaxAnalysis)
  private
    FScriptProcessor: TScriptProcessor;
    { Rules }
    procedure MapFieldList(AFieldsMap: TFieldsMap);
    procedure SetStatement;
    procedure ShowStatement;
    procedure CreateMap;
    procedure CreateStatement;
    procedure ExecuteStatement;
    procedure AlterStatement;
    procedure LogStatement;
    procedure CopyStatement;
    procedure SyncStatement;
    procedure StatementList;
    procedure MainProgram;
    procedure Statement;
  public
    function Execute: TValue; override;

    constructor Create(const AFileName: string; AScriptProcessor: TScriptProcessor); reintroduce;
    constructor CreateFromString(const AScript: string; AScriptProcessor: TScriptProcessor);
    destructor Destroy; override;
  end;

implementation

{ TRunner }

// AlterStatement ::= ALTER TRIGGER ( trigger_name | ALL ) (ACTIVE | INACTIVE)
procedure TRunner.AlterStatement;
var
  LTriggerName: string;
  LAction: string;
begin
  NextToken;

  ExpectText(TTokenKind.Identifier, 'TRIGGER');

  Expect(TTokenKind.Identifier, False);

  LTriggerName := FToken.StringValue;

  NextToken;

  Expect(TTokenKind.Identifier, False);

  LAction := FToken.StringValue;

  if FToken.Kind <> TTokenKind.Identifier then
    raise ESyntaxError.CreateFmt('Expected ACTIVE or INACTIVE but found [%s]', [LAction], FToken);

  NextToken;

  FScriptProcessor.AlterTriggerStatus(LTriggerName, LAction);
end;

// CopyStatement ::= COPY FROM TABELLA_SORGENTE TO TABELLA_DESTINAZIONE
procedure TRunner.CopyStatement;
var
  LSourceTable: string;
  LTargetTable: string;
  LFilter: string;
begin
  NextToken;
  ExpectText(TTokenKind.Identifier, 'FROM');

  if FToken.Kind <> TTokenKind.Identifier then
    raise ESyntaxError.CreateFmt('Expected source table name but found [%s]', [FToken.ToString], FToken);

  LSourceTable := FToken.StringValue;
  NextToken;

  ExpectText(TTokenKind.Identifier, 'TO');

  if FToken.Kind <> TTokenKind.Identifier then
    raise ESyntaxError.CreateFmt('Expected source table name but found [%s]', [FToken.ToString], FToken);

  LTargetTable := FToken.StringValue;
  NextToken;

  LFilter := '';
  if (FToken.Kind = TTokenKind.Identifier) and SameText(FToken.StringValue, 'FILTER') then
  begin
    NextToken;
    Expect(TTokenKind.StringLiteral, False);
    LFilter := FToken.StringValue;
    NextToken;
  end;

  FScriptProcessor.Copy(LSourceTable, LTargetTable, LFilter);
end;

constructor TRunner.Create(const AFileName: string; AScriptProcessor: TScriptProcessor);
begin
  inherited Create(TTokenScanner.CreateFromFile(AFileName, TEncoding.ANSI), True);
  FScriptProcessor := AScriptProcessor;
end;

constructor TRunner.CreateFromString(const AScript: string;
  AScriptProcessor: TScriptProcessor);
begin
  inherited Create(TTokenScanner.CreateFromString(AScript), True);
  FScriptProcessor := AScriptProcessor;
end;

// CreateMap ::= FOR MAP_NAME "(" MapFieldList ")"
procedure TRunner.CreateMap;
var
  LFieldsMap: TFieldsMap;
begin
  ExpectText(TTokenKind.Identifier, 'FOR');

  LFieldsMap := TFieldsMap.Create;
  try
    Expect(TTokenKind.Identifier, False);
    LFieldsMap.TableName := FToken.StringValue.ToUpper;
    NextToken;

    Expect(TTokenKind.LeftParenthesis);
    MapFieldList(LFieldsMap);
    Expect(TTokenKind.RightParenthesis);

    FScriptProcessor.SaveMap(LFieldsMap);

  except
    LFieldsMap.Free;
    raise;
  end;
end;

// CreateStatement ::= CREATE ( MapStatement | ... )
procedure TRunner.CreateStatement;
var
  LObjectType: string;
begin
  NextToken;

  Expect(TTokenKind.Identifier, False);
  LObjectType := FToken.StringValue;
  NextToken;

  if SameText(LObjectType, 'MAP') then
    CreateMap
  else
    raise ESyntaxError.CreateFmt('Expected CREATE MAP but found [%s]', [FToken.ToString], FToken);
end;

destructor TRunner.Destroy;
begin
  inherited;
end;

function TRunner.Execute: TValue;
begin
  NextToken;
  MainProgram;
end;

// ExecuteStatement ::= EXECUTE ( SQL 'sql command...' | QUERY 'query string...' )
procedure TRunner.ExecuteStatement;
var
  LSQLScript: string;
  LSQLQuery: string;
begin
  LSQLScript := '';
  LSQLQuery := '';

  NextToken;

  Expect(TTokenKind.Identifier, False);

  if SameText(FToken.StringValue, 'SQL') then
  begin
    NextToken;
    Expect(TTokenKind.StringLiteral, False);
    LSQLScript := FToken.StringValue;
  end
  else if SameText(FToken.StringValue, 'QUERY') then
  begin
    NextToken;
    Expect(TTokenKind.StringLiteral, False);
    LSQLQuery := FToken.StringValue;
  end
  else
    raise ESyntaxError.CreateFmt('Expected SQL or QUERY but found [%s]', [FToken.ToString], FToken);

  NextToken;

  if LSQLScript <> '' then
    FScriptProcessor.ExecuteScript(LSQLScript)
  else if LSQLQuery <> '' then
    FScriptProcessor.ExecuteQuery(LSQLQuery)
  else
    raise Exception.Create('ExecuteStatement');
end;

// SyncStatement ::= SYNC SEQUENCE GENERATOR_NAME WITH TABLE_NAME
procedure TRunner.SyncStatement;
var
  LGeneratorName: string;
  LTableNameAndField: string;
  LFieldNamePair: TArray<string>;
begin
  NextToken;

  ExpectText(TTokenKind.Identifier, 'SEQUENCE');

  if FToken.Kind <> TTokenKind.Identifier then
    raise ESyntaxError.CreateFmt('Expected generator name but found [%s]', [FToken.ToString], FToken);

  LGeneratorName := FToken.StringValue;
  NextToken;

  ExpectText(TTokenKind.Identifier, 'WITH');

  if FToken.Kind <> TTokenKind.Identifier then
    raise ESyntaxError.CreateFmt('Expected table name but found [%s]', [FToken.ToString], FToken);

  LTableNameAndField := FToken.StringValue;

  LFieldNamePair := LTableNameAndField.Split(['.']);

  if Length(LFieldNamePair) <> 2 then
    raise ESyntaxError.CreateFmt('Expected TABLE.FIELD but found [%s]', [LTableNameAndField], FToken);

  NextToken;

  FScriptProcessor.SyncGenerator(LGeneratorName, LFieldNamePair[0], LFieldNamePair[1]);
end;

// LogStatement ::= LOG [DEBUG | INFO | WARNING |ERROR ] "string"
procedure TRunner.LogStatement;
var
  LLogLevel: TLogLevel;
begin
  LLogLevel := llInfo;

  NextToken;

  // Verifica se c' il log level
  if FToken.Kind = TTokenKind.Identifier then
  begin
    if SameText(FToken.StringValue, 'DEBUG') then
      LLogLevel := llDebug
    else if SameText(FToken.StringValue, 'INFO') then
      LLogLevel := llInfo
    else if SameText(FToken.StringValue, 'WARNING') then
      LLogLevel := llWarning
    else if SameText(FToken.StringValue, 'ERROR') then
      LLogLevel := llError
    else
      raise ESyntaxError.CreateFmt('Expected DEBUG, INFO, WARNING or ERROR but found [%s]', [FToken.ToString], FToken);

    NextToken;
  end;

  if FToken.Kind <> TTokenKind.StringLiteral then
    raise ESyntaxError.CreateFmt('Expected string but found [%s]', [FToken.ToString], FToken);

  TLogger.Log(LLogLevel, FToken.StringValue);
  NextToken;
end;

// MainProgram ::= StatementList , EOF
procedure TRunner.MainProgram;
begin
  StatementList;
  Expect(TTokenKind.EndOfStream);
end;

// MapFieldList ::= identifier AS identifier [ , identifier AS identifier [ , identifier AS identifier ...] ]
procedure TRunner.MapFieldList(AFieldsMap: TFieldsMap);
var
  LMapFieldDef: TMapFieldDef;
begin
  LMapFieldDef := TMapFieldDef.Create;
  try
    Expect(TTokenKind.Identifier, False);
    LMapFieldDef.SourceFieldName := FToken.StringValue.ToUpper;
    NextToken;

    ExpectText(TTokenKind.Identifier, 'AS');

    Expect(TTokenKind.Identifier, False);
    LMapFieldDef.DestinationFieldName := FToken.StringValue.ToUpper;
    NextToken;

    AFieldsMap.MapFields.Add(LMapFieldDef.SourceFieldName, LMapFieldDef);
  except
    LMapFieldDef.Free;
    raise;
  end;

  if FToken.Kind = TTokenKind.Comma then
  begin
    NextToken;
    MapFieldList(AFieldsMap);
  end;

end;

// SetStatement ::= SET VARIABLE_NAME "value"
procedure TRunner.SetStatement;
var
  LVariableName: string;
  LValue: Variant;
begin
  NextToken;

  Expect(TTokenKind.Identifier, False);
  LVariableName := FToken.StringValue;
  NextToken;

  if FToken.Kind = TTokenKind.StringLiteral then
    LValue := FToken.StringValue
  else if FToken.Kind = TTokenKind.IntegerLiteral then
    LValue := FToken.IntegerValue
  else if FToken.Kind = TTokenKind.FloatLiteral then
    LValue := FToken.FloatValue
  else if FToken.Kind = TTokenKind.KeywordFalse then
    LValue := False
  else if FToken.Kind = TTokenKind.KeywordTrue then
    LValue := True
  else if FToken.Kind = TTokenKind.KeywordNull then
    LValue := Null
  else
    raise ESyntaxError.CreateFmt('Expected literal value but found [%s]', [FToken.ToString], FToken);

  NextToken;

  FScriptProcessor.SetVariable(LVariableName, LValue);
end;

// ShowStatement ::= SHOW VARIABLE_NAME
procedure TRunner.ShowStatement;
var
  LVariableName: string;
  LValue: Variant;
begin
  NextToken;

  Expect(TTokenKind.Identifier, False);

  LVariableName := FToken.StringValue;

  NextToken;

  LValue := FScriptProcessor.GetVariable(LVariableName);

  FScriptProcessor.DisplayText(VariableToStr(LValue));

end;

// Statement ::= CopyStatement | SyncStatement | AlterStatement | CreateStatement
procedure TRunner.Statement;
begin
  Expect(TTokenKind.Identifier, False);
  if SameText(FToken.StringValue, 'COPY') then
    CopyStatement
  else if SameText(FToken.StringValue, 'SYNC') then
    SyncStatement
  else if SameText(FToken.StringValue, 'LOG') then
    LogStatement
  else if SameText(FToken.StringValue, 'ALTER') then
    AlterStatement
  else if SameText(FToken.StringValue, 'CREATE') then
    CreateStatement
  else if SameText(FToken.StringValue, 'EXECUTE') then
    ExecuteStatement
  else if SameText(FToken.StringValue, 'SET') then
    SetStatement
  else if SameText(FToken.StringValue, 'SHOW') then
    ShowStatement
  else
    raise ESyntaxError.CreateFmt('Unrecognized statement [%s]', [FToken.ToString], FToken);
end;

// StatementList ::= Statement { ';' , Statement }
procedure TRunner.StatementList;
begin
  Statement;
  while FToken.Kind = TTokenKind.Semicolon do
  begin
    Expect(TTokenKind.Semicolon);
    if FToken.Kind in [TTokenKind.EndOfStream] then
      Break;
    Statement;
  end;
end;

end.
