unit DBBuilder.ISCScript.Processor;

interface

uses
  System.SysUtils, System.Classes, Data.DB, DBBuilder.DB.DatabaseManager,
  System.Generics.Collections, System.Variants, DBBuilder.DB.Core;

type
  EScriptProcessorError = class(Exception);

  TMapFieldDef = class(TObject)
  private
    FSourceFieldName: string;
    FDestinationFieldName: string;
  public
    property SourceFieldName: string read FSourceFieldName write FSourceFieldName;
    property DestinationFieldName: string read FDestinationFieldName write FDestinationFieldName;
  end;

  TMapFields = class(TObjectDictionary<string, TMapFieldDef>)
  end;

  TFieldsMap = class(TObject)
  private
    FTableName: string;
    FMapFields: TMapFields;
  public
    function GetDestinationFieldName(const ASourceFieldName: string): string;
    property TableName: string read FTableName write FTableName;
    property MapFields: TMapFields read FMapFields;

    constructor Create;
    destructor Destroy; override;
  end;

  TFieldsMaps = class(TObjectDictionary<string, TFieldsMap>)
  end;

  TVariables = class(TDictionary<string,Variant>)
  end;

  TScriptProcessor = class(TObject)
  private
    FDefaultConnection: TDatabaseManager;
    FFieldsMaps: TFieldsMaps;
    FVariables: TVariables;
    procedure FillParams(ADataSet: TDataSet; AParams: TDBParams; AMap: TFieldsMap);
    //function RecordToInsert(ADataSet: TDataSet; const ATableName: string; ADatabaseManager: TDatabaseManager; AMap: TFieldsMap): string;
    function GetInsertStatement(ADataSet: TDataSet; const ATableName: string; ADatabaseManager: TDatabaseManager; AMap: TFieldsMap): string;
    function DestinationFieldName(const ASourceFieldName: string; LMap: TFieldsMap): string;
    procedure SetDefaultConnection(const Value: TDatabaseManager);
  public
    property DefaultConnection: TDatabaseManager read FDefaultConnection write SetDefaultConnection;

    procedure ClearMaps;
    procedure Copy(const ASourceTable, ATargetTable, AFilter: string);
    procedure SyncGenerator(const AGeneratorName, ATableName, AFieldName: string);
    procedure AlterTriggerStatus(const ATriggerName, AAction: string);
    procedure SaveMap(AFieldsMap: TFieldsMap);
    procedure ExecuteScript(const AScript: string);
    procedure ExecuteQuery(const AQuery: string);
    procedure SetVariable(const AName: string; AValue: Variant);
    function GetVariable(const AName: string): Variant;
    procedure DisplayText(const AText: string);

    constructor Create;
    destructor Destroy; override;
  end;

function DefaultScriptProcessor: TScriptProcessor;

function VariableToStr(AValue: Variant): string;

implementation

{ TScriptProcessor }

uses
  DBBuilder.Core.Logging;

var
  _DefaultScriptProcessor: TScriptProcessor;

const
  // Predefined variables
  DEFAULT_CONNECTION = 'DEFAULT_CONNECTION';
  SOURCE_DATABASE = 'SOURCE_DATABASE';
  TARGET_DATABASE = 'TARGET_DATABASE';
  EXE_NAME = 'EXE_NAME';

function DefaultScriptProcessor: TScriptProcessor;
begin
  if not Assigned(_DefaultScriptProcessor) then
    _DefaultScriptProcessor := TScriptProcessor.Create;

  Result := _DefaultScriptProcessor;
end;

function VariableToStr(AValue: Variant): string;
begin
  if VarIsNull(AValue) then
    Result := 'NULL'
  else
    Result := VarToStr(AValue);
end;

procedure TScriptProcessor.AlterTriggerStatus(const ATriggerName,
  AAction: string);
var
  LTriggerStatus: TTriggerStatus;
begin
  if SameText(AAction, 'ACTIVE') then
    LTriggerStatus := tsActive
  else if SameText(AAction, 'INACTIVE') then
    LTriggerStatus := tsInactive
  else
    raise EScriptProcessorError.CreateFmt('Stato "%s" non valido', [AAction]);

  if SameText(ATriggerName, 'ALL') then
    FDefaultConnection.AlterAllTriggerStatus(LTriggerStatus)
  else
    FDefaultConnection.AlterTriggerStatus(ATriggerName, LTriggerStatus);
end;

procedure TScriptProcessor.ClearMaps;
begin
  FFieldsMaps.Clear;
end;

procedure TScriptProcessor.Copy(const ASourceTable, ATargetTable, AFilter: string);
var
  LSourceDataSet: TDataSet;
  LInsertStatement: string;
  LWhereCondition: string;
  LMap: TFieldsMap;
  LParams: TDBParams;
begin
  if not FFieldsMaps.TryGetValue(ATargetTable, LMap) then
    LMap := nil;

  if AFilter = '' then
    LWhereCondition := ''
  else
    LWhereCondition := ' WHERE ' + AFilter;

  LParams := TDBParams.Create;
  try
    try
      LSourceDataSet := SourceDatabaseManager.CreateDataSet('SELECT * FROM ' + ASourceTable + LWhereCondition);
      try
        TLogger.Log(llInfo, 'Trovate %d righe', [LSourceDataSet.RecordCount]);
        LSourceDataSet.First;
        LInsertStatement := GetInsertStatement(LSourceDataSet, ATargetTable, TargetDatabaseManager, LMap);
        while not LSourceDataSet.Eof do
        begin
          LParams.Clear;
          FillParams(LSourceDataSet, LParams, LMap);
          TLogger.Log(llDebug, LInsertStatement);
          TargetDatabaseManager.ExecuteStatement(LInsertStatement, LParams);
          LSourceDataSet.Next;
        end;
      finally
        LSourceDataSet.Free;
      end;
      TargetDatabaseManager.Commit;
    except
      TargetDatabaseManager.Rollback;
      raise;
    end;
  finally
    LParams.Free;
  end;
end;

constructor TScriptProcessor.Create;
begin
  FFieldsMaps := TFieldsMaps.Create([doOwnsValues]);
  FVariables := TVariables.Create;
  FDefaultConnection := TargetDatabaseManager;
end;

function TScriptProcessor.DestinationFieldName(const ASourceFieldName: string;
  LMap: TFieldsMap): string;
begin
  if not Assigned(LMap) then
    Exit(ASourceFieldName);

  Result := LMap.GetDestinationFieldName(ASourceFieldName);
end;

destructor TScriptProcessor.Destroy;
begin
  FFieldsMaps.Free;
  FVariables.Free;
  inherited;
end;

procedure TScriptProcessor.DisplayText(const AText: string);
begin
  TLogger.Log(llInfo, AText);
end;

procedure TScriptProcessor.ExecuteQuery(const AQuery: string);
var
  LDataSet: TDataSet;
  LIndex: Integer;
  LLine: string;
  LText: string;
begin
  LDataSet := FDefaultConnection.CreateDataSet(AQuery);
  try
    LLine := '';
    LText := '';
    for LIndex := 0 to LDataSet.FieldCount - 1 do
      LLine := LLine + LDataSet.Fields[LIndex].FieldName + ' | ';

    LText := LText + LLine + sLineBreak;

    while not LDataSet.Eof do
    begin
      LLine := '';
      for LIndex := 0 to LDataSet.FieldCount - 1 do
        LLine := LLine + LDataSet.Fields[LIndex].Text + ' | ';
      LText := LText + LLine + sLineBreak;

      LDataSet.Next;
    end;

    DisplayText(LText);

  finally
    LDataSet.Free;
  end;
end;

procedure TScriptProcessor.ExecuteScript(const AScript: string);
begin
  FDefaultConnection.ExecuteScript(AScript);
end;

procedure TScriptProcessor.FillParams(ADataSet: TDataSet; AParams: TDBParams; AMap: TFieldsMap);
var
  LIndex: Integer;
  LField: TField;
  LDestinationFieldName: string;
begin
  for LIndex := 0 to ADataSet.FieldCount - 1 do
  begin
    LField := ADataSet.Fields[LIndex];
    LDestinationFieldName := DestinationFieldName(LField.FieldName, AMap);
    if LDestinationFieldName <> '' then
      AParams.Add(LDestinationFieldName, LField.DataType, LField.Value);
  end;
end;

function TScriptProcessor.GetInsertStatement(ADataSet: TDataSet;
  const ATableName: string; ADatabaseManager: TDatabaseManager;
  AMap: TFieldsMap): string;
var
  LIndex: Integer;
  LFields: string;
  LValues: string;
  LDestinationFieldName: string;
begin
  LFields := '';
  LValues := '';
  for LIndex := 0 to ADataSet.FieldCount - 1 do
  begin
    LDestinationFieldName := DestinationFieldName(ADataSet.Fields[LIndex].FieldName, AMap);
    if LDestinationFieldName <> '' then
    begin
      if LFields <> '' then
      begin
        LFields := LFields + ', ';
        LValues := LValues + ', ';
      end;
      LFields := LFields + LDestinationFieldName;
      LValues := LValues + ':' + LDestinationFieldName;
      //LValues := LValues + ADatabaseManager.QuoteValue(ADataSet.Fields[LIndex].Value, ADataSet.Fields[LIndex].DataType);
    end;
  end;
  Result := 'INSERT INTO ' + ATableName + ' (' + LFields + ') VALUES (' + LValues + ')';
end;

function TScriptProcessor.GetVariable(const AName: string): Variant;
begin
  if SameText(AName, DEFAULT_CONNECTION) then
  begin
    if FDefaultConnection = TargetDatabaseManager then
      Result := 'TARGET'
    else if FDefaultConnection = SourceDatabaseManager then
      Result := 'SOURCE'
    else
      Result := Null;
  end
  else if SameText(AName, SOURCE_DATABASE) then
  begin
    Result := SourceDatabaseManager.ConnectionString;
  end
  else if SameText(AName, TARGET_DATABASE) then
  begin
    Result := TargetDatabaseManager.ConnectionString;
  end
  else if SameText(AName, EXE_NAME) then
  begin
    Result := ExtractFileName(ParamStr(0));
  end
  else if not FVariables.TryGetValue(AName, Result) then
    Result := Null;
end;

//function TScriptProcessor.RecordToInsert(ADataSet: TDataSet;
//  const ATableName: string; ADatabaseManager: TDatabaseManager;
//  AMap: TFieldsMap): string;
//var
//  LIndex: Integer;
//  LFields: string;
//  LValues: string;
//begin
//  LFields := '';
//  LValues := '';
//  for LIndex := 0 to ADataSet.FieldCount - 1 do
//  begin
//    LFields := LFields + DestinationFieldName(ADataSet.Fields[LIndex].FieldName, AMap);
//    LValues := LValues + ADatabaseManager.QuoteValue(ADataSet.Fields[LIndex].Value, ADataSet.Fields[LIndex].DataType);
//    if LIndex < ADataSet.FieldCount - 1 then
//    begin
//      LFields := LFields + ', ';
//      LValues := LValues + ', ';
//    end;
//  end;
//  Result := 'INSERT INTO ' + ATableName + ' (' + LFields + ') VALUES (' + LValues + ')';
//end;

procedure TScriptProcessor.SaveMap(AFieldsMap: TFieldsMap);
var
  LMapFieldDefPair: TPair<string,TMapFieldDef>;
begin
  if FFieldsMaps.ContainsKey(AFieldsMap.TableName) then
  begin
    TLogger.Log(llWarning, 'Mappa per tabella "%s" gi presente. Aggiorno la mappa con la nuova definizione', [AFieldsMap.TableName]);
    FFieldsMaps.Remove(AFieldsMap.TableName);
  end;
  //raise EScriptProcessorError.CreateFmt('Mappa per tabella "%s" gi presente', [AFieldsMap.TableName]);

  FFieldsMaps.Add(AFieldsMap.TableName, AFieldsMap);

  TLogger.Log(llInfo, 'Salvata mappa per tabella: "%s"', [AFieldsMap.TableName]);
  for LMapFieldDefPair in AFieldsMap.MapFields do
    TLogger.Log(llDebug, '  "%s" => "%s"', [LMapFieldDefPair.Value.SourceFieldName, LMapFieldDefPair.Value.DestinationFieldName]);
end;

procedure TScriptProcessor.SetDefaultConnection(const Value: TDatabaseManager);
begin
  if FDefaultConnection <> Value then
  begin
    FDefaultConnection := Value;
    TLogger.Log(llInfo, 'Connessione impostata su "%s"', [FDefaultConnection.Name]);
  end;
end;

procedure TScriptProcessor.SetVariable(const AName: string; AValue: Variant);
begin
  if SameText(AName, DEFAULT_CONNECTION) then
  begin
    if SameText(VarToStr(AValue), 'SOURCE') then
      DefaultConnection := SourceDatabaseManager
    else if SameText(VarToStr(AValue), 'TARGET') then
      DefaultConnection := TargetDatabaseManager
    else
      raise EScriptProcessorError.CreateFmt('La variabile "%s" pu assumere solo i valori "SOURCE" o "TARGET"', [DEFAULT_CONNECTION]);
  end
  else if SameText(AName, SOURCE_DATABASE) then
    raise EScriptProcessorError.CreateFmt('La variabile "%s"  a sola lettura', [SOURCE_DATABASE])
  else if SameText(AName, TARGET_DATABASE) then
    raise EScriptProcessorError.CreateFmt('La variabile "%s"  a sola lettura', [TARGET_DATABASE])
  else if SameText(AName, EXE_NAME) then
    raise EScriptProcessorError.CreateFmt('La variabile "%s"  a sola lettura', [EXE_NAME])
  else
    FVariables.AddOrSetValue(AName, AValue);
end;

procedure TScriptProcessor.SyncGenerator(const AGeneratorName,
  ATableName, AFieldName: string);
var
  LDataSet: TDataSet;
  LValue: Int64;
begin
  TLogger.Log(llDebug, 'Sincronizzo generatore "%s" con "%s.%s"', [AGeneratorName, ATableName, AFieldName]);

  LDataSet := FDefaultConnection.CreateDataSet('SELECT MAX(' + AFieldName + ') FROM ' + ATableName);
  try
    if LDataSet.IsEmpty then
      raise EScriptProcessorError.CreateFmt('Nessun dato trovato per "%s"', [ATableName]);

    LValue := LDataSet.Fields[0].AsLargeInt;
  finally
    LDataSet.Free;
  end;

  FDefaultConnection.SetSequence(AGeneratorName, LValue);

end;

{ TFieldsMap }

constructor TFieldsMap.Create;
begin
  inherited;
  FMapFields := TMapFields.Create([doOwnsValues]);
end;

destructor TFieldsMap.Destroy;
begin
  FMapFields.Free;
  inherited;
end;

function TFieldsMap.GetDestinationFieldName(
  const ASourceFieldName: string): string;
var
  LMapFieldDef: TMapFieldDef;
begin
  if not FMapFields.TryGetValue(ASourceFieldName.ToUpper, LMapFieldDef) then
    Exit('');
  Result := LMapFieldDef.DestinationFieldName;
end;

initialization

  _DefaultScriptProcessor := nil;

finalization

  _DefaultScriptProcessor.Free;

end.
